from .arm import Arm
from .calf import Calf
from .references import References
from .thigh import Thigh
from .trunk import Trunk

__all__ = [
    'Arm',
    'Calf',
    'Thigh',
    'Trunk',
    'References',
]



import logging
from dataclasses import dataclass

import pandas as pd

from .references import References
from .sensor import Sensor

logger = logging.getLogger(__name__)


@dataclass
class Calf(Sensor):
    def get_kneeling(self, df: pd.DataFrame) -> pd.Series:
        kneel = (
            ~df['non-wear']
            & (df['inclination'] >= 84)
            & (df['direction'] > 45)
            & (df['thigh_direction'] > -20)
            & (df['thigh_side_tilt'].abs() < 30)
            & (~df['activity'].isin(['non-wear', 'lie']))
        )

        return kneel

    def get_squatting(self, df: pd.DataFrame) -> pd.Series:
        squat = (
            ~df['non-wear']
            & (df['inclination'].between(32, 84, inclusive='neither'))
            & (df['inclination'] >= -1.5 * df['thigh_inclination'] + 187)
            & (df['direction'] > 0)
            & (~df['activity'].isin(['non-wear', 'lie']))
        )
        return squat

    def compute_activities(
        self,
        df: pd.DataFrame,
        activities: pd.DataFrame,
    ) -> pd.DataFrame:
        activities = activities.copy()
        df = df.copy()

        activities['activity'] = activities['activity'].cat.add_categories(['kneel', 'squat'])
        references = References()

        # Only keep overlapping data (thigh and calf)
        df = df.join(activities, how='inner')

        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)
        df['non-wear'] = self.get_non_wear(df)
        bouts = references.get_bouts(df['non-wear'])

        if self.orientation:
            df = self.fix_bouts_orientation(df, bouts)

        df['kneel'] = self.get_kneeling(df)
        df['squat'] = self.get_squatting(df)

        df.loc[df['kneel'], 'activity'] = 'kneel'
        df.loc[df['squat'], 'activity'] = 'squat'

        activities.loc[activities.index.isin(df.index), 'activity'] = df['activity']

        return activities

    def check_inside_out_flip(self, df: pd.DataFrame) -> bool:
        return False

    def rotate_by_reference_angle(self):
        raise NotImplementedError('Calf does not support rotating by reference angle.')

    def calculate_reference_angle(self):
        raise NotImplementedError('Calf does not support calculating reference angles.')



from .activities import Activities
from .classifications import References
from .exposures import Exposures
from .features import Features
from .sens_parser import Sens

__all__ = [
    'Exposures',
    'Features',
    'Activities',
    'References',
    'Sens',
]



import logging
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, Literal

import numpy as np
import pandas as pd

from .classifications import Arm, Calf, References, Thigh, Trunk
from .iterators import DataFrameIterator
from .settings import ACTIVITIES, CONFIG, FEATURES, SENS__ACTIVITY_VALUES, SENS__FLOAT_FACTOR

logger = logging.getLogger(__name__)


@dataclass
class Activities:
    system_frequency: int = 30
    vendor: Literal['Sens', 'Other'] = 'Other'
    orientation: bool = True
    chunks: bool = False
    size: str | timedelta = '1d'
    overlap: str | timedelta = '15min'
    config: dict[str, Any] | None = None

    def __post_init__(self):
        if isinstance(self.size, str):
            self.size = pd.Timedelta(self.size).to_pytimedelta()

        if isinstance(self.overlap, str):
            self.overlap = pd.Timedelta(self.overlap).to_pytimedelta()

        if self.config is None:
            self.config = CONFIG

    def _compute_chunk(
        self,
        thigh: pd.DataFrame,
        *,
        trunk: pd.DataFrame | None = None,
        calf: pd.DataFrame | None = None,
        arm: pd.DataFrame | None = None,
        references: References,
    ) -> tuple[pd.DataFrame, References]:
        not_overlaps = thigh[~thigh['overlap']]
        start, end = not_overlaps.index[0], not_overlaps.index[-1]

        if trunk is not None:
            trunk = trunk.loc[(trunk.index >= start) & (trunk.index < end)]

        if calf is not None:
            calf = calf.loc[(calf.index >= start) & (calf.index < end)]

        if arm is not None:
            arm = arm.loc[(arm.index >= start) & (arm.index < end)]

        df, references = self._compute(
            thigh=thigh,
            trunk=trunk,
            calf=calf,
            arm=arm,
            references=references,
        )

        df = df.loc[(df.index >= start) & (df.index < end)]

        return df, references

    def _compute_chunks(
        self,
        thigh: pd.DataFrame,
        *,
        trunk: pd.DataFrame | None = None,
        calf: pd.DataFrame | None = None,
        arm: pd.DataFrame | None = None,
        references: References,
    ) -> tuple[pd.DataFrame, References]:
        chunks = DataFrameIterator(thigh, size=self.size, overlap=self.overlap)

        output = []

        for chunk in chunks:
            df, references = self._compute_chunk(
                chunk,
                trunk=trunk,
                calf=calf,
                arm=arm,
                references=references,
            )
            output.append(df)

        output = pd.concat(output)
        output.sort_index(inplace=True)

        return output, references

    def _compute(
        self,
        thigh: pd.DataFrame,
        references: References,
        *,
        trunk: pd.DataFrame | None = None,
        calf: pd.DataFrame | None = None,
        arm: pd.DataFrame | None = None,
    ) -> tuple[pd.DataFrame, References]:
        if thigh.empty:
            raise ValueError('Thigh data is empty. Please provide valid thigh data.')

        activities = Thigh(
            orientation=self.orientation,
            system_frequency=self.system_frequency,
            vendor=self.vendor,
            config=self.config,
        ).compute_activities(thigh, references=references)
        logger.info('Detected activities for thigh.')

        if isinstance(trunk, pd.DataFrame) and not trunk.empty:
            activities = Trunk(orientation=self.orientation, config=self.config).compute_activities(
                trunk, activities, references=references
            )
            logger.info('Detected activities for trunk.')

        if isinstance(calf, pd.DataFrame) and not calf.empty:
            activities = Calf(orientation=self.orientation).compute_activities(calf, activities)
            logger.info('Detected activities for calf.')

        if isinstance(arm, pd.DataFrame) and not arm.empty:
            activities = activities.join(Arm(orientation=self.orientation).compute_activities(arm), how='left')
            logger.info('Detected activities for arm.')

        references.remove_outdated(activities.index[-1])

        return activities, references

    def compute(
        self,
        thigh: pd.DataFrame,
        *,
        trunk: pd.DataFrame | None = None,
        calf: pd.DataFrame | None = None,
        arm: pd.DataFrame | None = None,
        references: dict[str, Any] | None = None,
    ) -> tuple[pd.DataFrame, dict[str, Any]]:
        references_obj = References.from_dict(references)
        references_obj.remove_outdated(thigh.index[0])

        if self.chunks:
            activities, references_obj = self._compute_chunks(
                thigh=thigh,
                trunk=trunk,
                calf=calf,
                arm=arm,
                references=references_obj,
            )
        else:
            activities, references_obj = self._compute(
                thigh=thigh,
                trunk=trunk,
                calf=calf,
                arm=arm,
                references=references_obj,
            )

        return activities, references_obj.to_dict()

    def _map_activities(
        self,
        series: pd.Series,
        input: Literal['text', 'numeric'] = 'text',
    ) -> pd.Series:
        match input:
            case 'text':
                code_to_activity = {v: k for k, v in ACTIVITIES.items()}
                series = series.map(code_to_activity)
            case 'numeric':
                series = series.map(ACTIVITIES)
            case _:
                raise ValueError("Input must be either 'text' or 'numeric'.")

        if series.isna().all():
            raise ValueError('No valid activities found in the series. Please check the input data.')

        return series

    def to_sens(
        self,
        df: pd.DataFrame,
        references: dict[str, Any],
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, dict[str, Any]]:
        values = SENS__ACTIVITY_VALUES
        origin = df
        df = pd.DataFrame(index=origin.index)

        # "activity" is not part of SENS__ACTIVITY_VALUES, because it is always present in the dataframe
        for value in values:
            if value not in origin.columns:
                df[value] = 0  # If column is missing, fill with 0. Usually because not all sensors are present
            else:
                df[value] = origin[value]

        df[values] = df[values].fillna(0)  # Fill NaN values with 0
        df['activity'] = self._map_activities(origin['activity'], 'text').astype(np.int16)  # From the origin dataframe

        df[values] = (df[values] * SENS__FLOAT_FACTOR).astype(np.int32)
        df['verbose'] = 0
        df['verbose'] = df['verbose'].astype(np.int32)

        df.index = df.index.astype(np.int64) // 10**6  # Time in milliseconds

        return (
            df.index.values,
            df['activity'].values,
            df[values].values,
            df['verbose'].values,
            references,
        )  # type: ignore

    def _parse_sensor_features(
        self,
        timestamps: np.ndarray,
        data: np.ndarray,
    ) -> pd.DataFrame:
        df = pd.DataFrame(data, index=timestamps, columns=FEATURES)
        df.index = pd.to_datetime(df.index, unit='ms')  # type: ignore
        df.index.name = 'datetime'  # type: ignore

        df[FEATURES] = (df[FEATURES] / SENS__FLOAT_FACTOR).astype(np.float32)

        return df

    def _raw_from_sens(
        self,
        timestamps: np.ndarray,
        data: np.ndarray,
    ) -> tuple[pd.DataFrame, pd.DataFrame | None, pd.DataFrame | None, pd.DataFrame | None]:
        thigh, trunk, calf, arm = None, None, None, None

        thigh = self._parse_sensor_features(timestamps[0], data[0][:, 1:])

        if timestamps[1] is not None and data[1] is not None:
            trunk = self._parse_sensor_features(timestamps[1], data[1][:, 1:])

        if timestamps[2] is not None and data[2] is not None:
            calf = self._parse_sensor_features(timestamps[2], data[2][:, 1:])

        if timestamps[3] is not None and data[3] is not None:
            arm = self._parse_sensor_features(timestamps[3], data[3][:, 1:])

        return thigh, trunk, calf, arm

    def compute_sens(
        self,
        timestamps: np.ndarray,
        data: np.ndarray,
        references: dict[str, Any] | None = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, dict[str, Any]]:
        thigh, trunk, calf, arm = self._raw_from_sens(timestamps, data)

        df, references = self.compute(thigh=thigh, trunk=trunk, calf=calf, arm=arm, references=references)

        return self.to_sens(df, references)



import logging
from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd

from .references import References
from .sensor import Calculation, Sensor

logger = logging.getLogger(__name__)


@dataclass
class Trunk(Sensor):
    config: dict[str, Any]

    def _reference_angle_string(self, angle: np.ndarray) -> str:
        return f'[{angle[0]:.2f}, {angle[1]:.2f}, {angle[2]:.2f}]'

    def calculate_reference_angle(self, df: pd.DataFrame) -> tuple[np.ndarray, Calculation]:
        default_angle = np.array([27, 27, 0])
        angle_status = Calculation.DEFAULT

        walk = df[(~df['non-wear']) & (df['activity'] == 'walk')]

        if not walk.empty:
            y = np.median(walk['direction']) - 6
            z = np.median(walk['side_tilt'])
            x = np.degrees(np.arccos(np.radians(y)) * np.cos(np.radians(z)))

            reference_angle = np.array([x, y, z])
            angle_status = Calculation.AUTOMATIC

            logger.info(f'Reference angle calculated: {self._reference_angle_string(reference_angle)} degrees.')
        else:
            reference_angle = default_angle
            logger.info(
                f'No walking data available. Using default reference angle: {self._reference_angle_string(reference_angle)} degrees.'
            )

        return np.float32(np.radians(reference_angle)), angle_status  # type: ignore

    def rotate_by_reference_angle(self, df: pd.DataFrame, angle: np.ndarray) -> pd.DataFrame:  # type: ignore
        df = df.copy()
        x, y, z = angle

        cos_y, sin_y = np.cos(y), np.sin(y)
        cos_z, sin_z = np.cos(z), np.sin(z)

        rotation_y = np.array(
            [
                [cos_y, 0, sin_y],
                [0, 1, 0],
                [-sin_y, 0, cos_y],
            ]
        )
        rotation_z = np.array(
            [
                [cos_z, sin_z, 0],
                [-sin_z, cos_z, 0],
                [0, 0, 1],
            ]
        )
        rotation_matrix = np.matmul(rotation_y, rotation_z)  # type: np.ndarray

        df[['x', 'y', 'z']] = df[['x', 'y', 'z']].dot(rotation_matrix).astype(np.float32)
        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)

        return df

    def check_inside_out_flip(self, df: pd.DataFrame) -> bool:
        valid_points = df[(df['inclination'] < 45) | (df['inclination'] > 135)]  # Standing

        if valid_points.empty:
            logger.warning('Not enough data to check inside out flip. Skipping.')
            return False

        mdn = np.median(valid_points['direction'])
        flip = False if (mdn > 0.1) else True

        if flip:
            logger.warning(f'Inside out flip detected (median z: {mdn:.2f}).')

        return flip

    def get_backwards(
        self,
        df: pd.DataFrame,
        inclination_angle: float,
        **kwargs,
    ) -> pd.Series:
        backwards = (df['direction'] < -inclination_angle) | (df['side_tilt'].abs() > inclination_angle)
        backwards.name = 'backwards'

        return backwards

    def fix_lie(
        self,
        df: pd.DataFrame,
        orientation_angle: float,
        **kwargs,
    ) -> pd.Series:
        valid = df[['activity']].copy()
        mask = ~df['non-wear']  # Only consider wear time
        valid.loc[mask & (valid['activity'] == 'lie'), 'activity'] = 'sit'  # Change every lie to sit
        valid.loc[mask & (valid['activity'] == 'sit') & (df['inclination'] > orientation_angle), 'activity'] = 'lie'
        valid.loc[mask & (valid['activity'] == 'sit') & (df['backwards']), 'activity'] = 'lie'

        return valid['activity']

    def fix_sit(
        self,
        df: pd.DataFrame,
        orientation_angle: float,
        inclination_angle: float,
        **kwargs,
    ) -> pd.Series:
        valid = df[['activity']].copy()
        mask = (~df['non-wear']) & (valid['activity'] == 'lie')  # Only consider wear time and lie activity
        valid.loc[mask & (df['direction'] > 0) & (df['thigh_direction'] > inclination_angle), 'activity'] = 'sit'

        valid.loc[
            mask
            & ((df['inclination'] - df['direction']).abs() < 10)
            & (df['inclination'] < orientation_angle)
            & (df['direction'] < orientation_angle),
            'activity',
        ] = 'sit'

        return valid['activity']

    def compute_activities(
        self,
        df: pd.DataFrame,
        activities: pd.DataFrame,
        references: References,
    ) -> pd.DataFrame:
        thigh_config = self.config['thigh']
        config = self.config['trunk']
        bouts_length = {activity[0]: activity[1]['bout'] for activity in thigh_config.items()}

        activities = activities.copy()
        df = df.copy()

        # Only keep overlapping data (thigh and trunk)
        df = df.join(activities, how='inner')

        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)
        df['non-wear'] = self.get_non_wear(df)
        bouts = references.get_bouts(df['non-wear'], 'trunk')

        if self.orientation:
            df = self.fix_bouts_orientation(df, bouts)

        df, bouts = self.rotate_bouts_by_reference_angles(df, bouts)

        df['backwards'] = self.get_backwards(df, **config['lie'])
        df['activity'] = self.fix_lie(df, **config['lie'])
        df['activity'] = self.fix_sit(df, **config['lie'])

        for activity in ['sit', 'lie']:
            df['activity'] = self.fix_bouts(df['activity'], activity, bouts_length[activity])

        df.loc[df['non-wear'], 'inclination'] = np.nan
        df.loc[df['non-wear'], 'direction'] = np.nan

        df.rename(
            columns={
                'inclination': 'trunk_inclination',
                'side_tilt': 'trunk_side_tilt',
                'direction': 'trunk_direction',
            },
            inplace=True,
        )

        activities = activities.join(df[['trunk_inclination', 'trunk_side_tilt', 'trunk_direction']], how='left')
        activities.loc[activities.index.isin(df.index), 'activity'] = df['activity']

        references.update_angle(bouts, 'trunk')

        return activities



import logging
import math
from dataclasses import dataclass
from typing import Any, Literal

import numpy as np
import pandas as pd
from scipy.signal import medfilt

from .references import References
from .sensor import Calculation, Sensor

logger = logging.getLogger(__name__)


@dataclass
class Thigh(Sensor):
    system_frequency: int
    vendor: Literal['Sens', 'Other']
    config: dict[str, Any]
    # rotate: bool = False # TODO: Implement rotation logic first.

    def check_inside_out_flip(self, df: pd.DataFrame) -> bool:
        # TODO: Check that this works correctly.
        rows_per_hour = self.system_frequency * 60 * 2
        window = rows_per_hour * 3
        step = rows_per_hour
        min_periods = rows_per_hour

        inclination = df['inclination']
        z = df['z']

        outliers = inclination.rolling(window=window, step=step, min_periods=min_periods).quantile(0.02) > 45
        outliers_mask = pd.Series(df.index.map(outliers), index=df.index, dtype='boolean').ffill()
        valid_points = z.loc[~outliers_mask & (inclination > 45)]

        if valid_points.empty:
            logger.warning('Not enough data to check inside out flip. Skipping.')
            return False

        mdn = np.median(valid_points)

        flip = True if (mdn > 0.1) else False

        if flip:
            logger.warning(f'Inside out flip detected (median z: {mdn:.2f}).')

        return flip

    def calculate_reference_angle(self, df: pd.DataFrame) -> tuple[float, Calculation]:
        x_threshold_lower = 0.1
        x_threshold_upper = 0.72  # NOTE: Originally 0.7. To match the walk.py, 0.72 should be used.
        inclination_threshold = 45  # NOTE: Same as stationary_threshold for walking.
        direction_threshold = 10

        angle_mdn_coefficient = 6
        angle_threshold_lower = -30  # FIXME: In new code this is -30, original: -28
        angle_threshold_upper = 15
        default_angle = -16
        angle_status = Calculation.DEFAULT

        walk_mask = (
            (df['sd_x'].between(x_threshold_lower, x_threshold_upper, inclusive='neither'))
            & (df['inclination'] < inclination_threshold)
            & (df['direction'] < direction_threshold)
        )
        walk = df[walk_mask]

        if not walk.empty:
            reference_angle = (
                np.median(walk['direction']) - angle_mdn_coefficient
            ).item()  # Walk direction reference angle (median, degrees)

            reference_angle = reference_angle * 0.725 - 5.569  # Correction factor based on RAW data.

            if (reference_angle < angle_threshold_lower) or (reference_angle > angle_threshold_upper):
                reference_angle = default_angle
                logger.warning(
                    f'Reference angle {reference_angle:.2f} degrees is outside the threshold range. Using default reference angle: {reference_angle:.2f} degrees.'
                )

            else:
                angle_status = Calculation.AUTOMATIC
                logger.info(f'Reference angle calculated: {reference_angle:.2f} degrees.')
        else:
            reference_angle = default_angle
            logger.warning(f'No valid walk data found. Using default reference angle: {reference_angle:.2f} degrees.')

        return np.float32(np.radians(reference_angle)).item(), angle_status

    def _rotate_sd(self, df: pd.DataFrame, angle: float) -> pd.DataFrame:
        sin = np.sin(angle)
        cos = np.cos(angle)

        sq_sin = np.square(sin)
        sq_cos = np.square(cos)

        sq_x = np.square(df['x'])
        sq_z = np.square(df['z'])

        sd = pd.DataFrame(index=df.index)

        sd['terms_x'] = (
            (sq_sin * df['sq_sum_z'])
            + (sq_cos * df['sq_sum_x'])
            + (2 * self.system_frequency * sq_x)
            + (2 * sin * df['x'] * df['sum_z'])
            + (-2 * sin * cos * df['sum_dot_xz'])
            + (-2 * cos * df['x'] * df['sum_x'])
        )
        sd.loc[sd['terms_x'] <= 0, 'terms_x'] = 0
        sd['sd_x'] = np.sqrt(1 / (2 * self.system_frequency - 1) * sd['terms_x'])

        sd['terms_z'] = (
            (sq_sin * df['sq_sum_x'])
            + (sq_cos * df['sq_sum_z'])
            + (2 * self.system_frequency * sq_z)
            + (2 * sin * cos * df['sum_dot_xz'])
            + (-2 * sin * df['z'] * df['sum_x'])
            + (-2 * cos * df['z'] * df['sum_z'])
        )
        sd.loc[sd['terms_z'] <= 0, 'terms_z'] = 0
        sd['sd_z'] = np.sqrt(1 / (2 * self.system_frequency - 1) * sd['terms_z'])

        sd['sd_y'] = df['sd_y']

        return sd[['sd_x', 'sd_y', 'sd_z']].astype(np.float32)

    def rotate_by_reference_angle(self, df: pd.DataFrame, angle: float) -> pd.DataFrame:  # type: ignore
        df = df.copy()
        angle = np.float32(angle)  # type: ignore
        cos_angle = np.cos(angle)
        sin_angle = np.sin(angle)

        rotation_matrix = np.array(
            [
                [cos_angle, 0, sin_angle],
                [0, 1, 0],
                [-sin_angle, 0, cos_angle],
            ]
        )

        df[['x', 'y', 'z']] = df[['x', 'y', 'z']].dot(rotation_matrix).astype(np.float32)
        df[['sd_x', 'sd_y', 'sd_z']] = self._rotate_sd(df, angle)  # TODO: Maybe not needed

        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)

        return df

    def _downsample_sd_correction_for_sens(
        self,
        df: pd.DataFrame,
        sampling_frequency: float,
        tolerance: float = 1,
    ) -> None:
        values, correction = None, None

        if math.isclose(sampling_frequency, 25, abs_tol=tolerance):
            values = (0.18, 1.03)
            correction = 25

        elif math.isclose(sampling_frequency, 12.5, abs_tol=tolerance):
            values = (0.02, 1.14)
            correction = 12.5

        if values:
            df[['sd_x', 'sd_y', 'sd_z']] = df[['sd_x', 'sd_y', 'sd_z']].apply(
                lambda x: values[0] * x**2 + values[1] * x
            )
            logger.info(
                f'Applied SD correction [{values[0]} * x^2 + {values[1]} * x] which is specific for Sens sensors at {correction:.1f} Hz ({sampling_frequency:.2f} Hz).'
            )
        else:
            logger.warning(f'No correction applied for sampling frequency {sampling_frequency:.2f} Hz on axes.')

    def _median_filter(
        self,
        valid: pd.Series,
        bouts_length: int,
    ) -> pd.Series:
        length = 2 * bouts_length - 1
        filtered = medfilt(valid.astype(int), length)
        filtered = medfilt(filtered, length)

        valid = pd.Series(filtered.astype(bool), index=valid.index, name='median_filtered')

        return valid

    def get_row(
        self,
        df: pd.DataFrame,
        bout: int,
        movement_threshold: float,
        inclination_angle: float,
        **kwargs,
    ) -> pd.Series:
        # NOTE: Do we miss the "mean anterior-posterior angle is greater than 40°"?
        valid = (inclination_angle < df['inclination']) & (movement_threshold < df['sd_x'])
        valid = self._median_filter(valid, bout)
        valid.name = 'row'

        return valid

    def get_bicycle(
        self,
        df: pd.DataFrame,
        bout: int,
        movement_threshold: float,
        inclination_angle: float,
        anterior_posterior_angle: float,
        direction_threshold: float,
        **kwargs,
    ) -> pd.Series:
        valid = (
            ((anterior_posterior_angle) < df['direction'])
            & (df['inclination'] < inclination_angle)
            & (movement_threshold < df['sd_x'])
        )

        valid = pd.Series(medfilt(valid.astype(int), 9), index=df.index)
        valid = valid & ((df['hl_ratio'] < 0.5) | (df['direction'] > direction_threshold))

        valid = self._median_filter(valid, bout)
        valid.name = 'bicycle'

        return valid

    def _get_stairs_threshold(
        self,
        df: pd.DataFrame,
        run_threshold: float,
        anterior_posterior_angle: float,
        stairs_threshold: float,
    ) -> float:
        valid = df['sd_x'].between(0.25, run_threshold, inclusive='neither') & (
            df['direction'] < anterior_posterior_angle
        )

        valid = df.loc[valid, 'direction']

        # FIXME: Maybe get different threshold if no valid data is found.
        if valid.empty:
            logger.warning('No valid data found for stairs threshold calculation. Using default stairs threshold.')
            valid = stairs_threshold
        else:
            valid = stairs_threshold + np.median(valid)  # type: ignore
            valid = valid.item()

        return valid

    def get_stairs(
        self,
        df: pd.DataFrame,
        bout: int,
        inclination_angle: float,
        movement_threshold: float,
        run_threshold: float,
        direction_threshold: float,
        stairs_threshold: float,
        anterior_posterior_angle: float,
        **kwargs,
    ) -> tuple[pd.Series, float]:
        stairs_threshold = self._get_stairs_threshold(df, run_threshold, anterior_posterior_angle, stairs_threshold)

        valid = (
            (stairs_threshold < df['direction'])
            & (df['direction'] < direction_threshold)
            & (movement_threshold < df['sd_x'])
            & (df['sd_x'] < run_threshold)
            & (df['inclination'] < inclination_angle)
        )

        valid = self._median_filter(valid, bout)
        valid.name = 'stairs'

        return valid, stairs_threshold

    def get_run(
        self,
        df: pd.DataFrame,
        bout: int,
        inclination_angle: float,
        run_threshold: float,
        **kwargs,  # Because there is extra parameter, so it will be ignored.
    ) -> pd.Series:
        valid = (df['sd_x'] > run_threshold) & (df['inclination'] < inclination_angle)

        valid = self._median_filter(valid, bout)
        valid.name = 'run'

        return valid

    def get_walk(
        self,
        df: pd.DataFrame,
        bout: int,
        stairs_threshold: float,
        inclination_angle: float,
        movement_threshold: float,
        run_threshold: float,
        **kwargs,
    ) -> pd.Series:
        valid = (
            (movement_threshold < df['sd_x'])
            & (df['sd_x'] < run_threshold)
            & (df['direction'] < stairs_threshold)
            & (df['inclination'] < inclination_angle)
        )

        valid = self._median_filter(valid, bout)
        valid.name = 'walk'

        return valid

    def get_stand(
        self,
        df: pd.DataFrame,
        bout: int,
        inclination_angle: float,
        movement_threshold: float,
        **kwargs,
    ) -> pd.Series:
        sd_max = np.max(df[['sd_x', 'sd_y', 'sd_z']], axis=1)
        valid = (df['inclination'] < inclination_angle) & (sd_max < movement_threshold)

        valid = self._median_filter(valid, bout)
        valid.name = 'stand'

        return valid

    def get_sit(
        self,
        df: pd.DataFrame,
        bout: int,
        inclination_angle: float,
        **kwargs,
    ) -> pd.Series:
        valid = df['inclination'] > inclination_angle
        valid = self._median_filter(valid, bout)
        valid.name = 'sit'

        return valid

    def _get_activity_column(self, df: pd.DataFrame) -> pd.Series:
        # Order matters here
        categories = [
            'row',
            'bicycle',
            'stairs',
            'run',
            'walk',
            'stand',
            'sit',
        ]

        df = df[categories].copy()
        df['shuffle'] = True  # If nothing else, it is shuffle

        categories.append('shuffle')
        activity = df[categories].idxmax(axis=1)
        categories = categories + ['lie', 'non-wear']
        activity = activity.astype(pd.CategoricalDtype(categories=categories))
        activity.name = 'activity'

        return activity

    def _fix_activities_bouts(self, df: pd.DataFrame, bouts: dict[str, int]) -> pd.Series:
        activities = df['activity']

        # Order matters here - shuffle position changed
        for activity in [
            'row',
            'bicycle',
            'stairs',
            'run',
            'walk',
            'shuffle',
            'stand',
            'sit',
        ]:
            activities = self.fix_bouts(activities, activity, bouts[activity])

        return activities

    def _get_rotational_crossing_points(
        self,
        df: pd.DataFrame,
        orientation_threshold: float,
    ) -> pd.DataFrame:
        noise_margin = 0.05

        thigh_angle = np.arcsin(df['y'] / np.sqrt(np.square(df['y']) + np.square(df['z'])))  # type: pd.Series # type:ignore
        thigh_angle = np.absolute(np.degrees(thigh_angle))  # type: pd.Series # type:ignore

        low = (thigh_angle < orientation_threshold).diff()
        high = (thigh_angle >= orientation_threshold).diff()

        noise = thigh_angle.diff().abs()  # type: float # type: ignore
        noise = noise >= noise_margin

        low = low & noise
        high = high & noise

        return pd.DataFrame({'low': low, 'high': high}, index=df.index)

    def get_lie(
        self,
        df: pd.DataFrame,
        bout: int,
        orientation_angle: float,
        **kwargs,
    ) -> pd.Series:
        df = df[['activity', 'y', 'z']].copy()
        df[['low', 'high']] = self._get_rotational_crossing_points(df, orientation_angle)
        df['lie'] = pd.Series(False, index=df.index, dtype=bool)
        df['bout'] = (df['activity'] != df['activity'].shift()).cumsum()

        specific_activity = df.loc[
            df['activity'] == 'sit', 'bout'
        ]  # NOTE: Activity "sit" needs to be already in the activities dataframe.
        bout_sizes = specific_activity.value_counts()
        specific_bouts = bout_sizes[bout_sizes > bout].index.values  # FIXME: Try > or >=

        df['specific'] = False
        df.loc[df['bout'].isin(specific_bouts), 'specific'] = True

        updated_bouts = df[df['specific']].groupby('bout').aggregate({'low': 'any', 'high': 'any'})
        updated_bouts = updated_bouts[updated_bouts['low'] & updated_bouts['high']]
        df['lie'] = False
        df.loc[df['bout'].isin(updated_bouts.index), 'lie'] = True

        return df['lie']

    def get_steps(self, df: pd.DataFrame) -> pd.Series:
        df = df[['activity', 'walk_feature', 'run_feature']].copy()
        scale = self.system_frequency / 2 * np.linspace(0, 1, 256)

        df['steps'] = 0
        df.loc[df['activity'].isin(['walk', 'stairs']), 'steps'] = df['walk_feature']
        df.loc[df['activity'] == 'run', 'steps'] = df['run_feature']
        df['steps'] = scale[df['steps']]
        df['steps'] = medfilt(df['steps'], 3)

        return df['steps'].astype(np.float32)

    def compute_activities(
        self,
        df: pd.DataFrame,
        references: References,
    ) -> pd.DataFrame:
        config = self.config['thigh']

        df = df.copy()
        sf = df['sf'].mode().values[0].item()

        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)
        non_wear = self.get_non_wear(df)
        bouts = references.get_bouts(non_wear, 'thigh')

        if self.orientation:
            df = self.fix_bouts_orientation(df, bouts)

        df, bouts = self.rotate_bouts_by_reference_angles(df, bouts)

        if self.vendor.lower() == 'sens':
            self._downsample_sd_correction_for_sens(df, sf)

        df['row'] = self.get_row(df, **config['row'])
        df['bicycle'] = self.get_bicycle(df, **config['bicycle'])
        df['stairs'], stairs_threshold = self.get_stairs(df, **config['stairs'])
        df['run'] = self.get_run(df, **config['run'])
        df['walk'] = self.get_walk(df, **config['walk'], stairs_threshold=stairs_threshold)
        df['stand'] = self.get_stand(df, **config['stand'])
        df['sit'] = self.get_sit(df, **config['sit'])

        df['activity'] = self._get_activity_column(df)
        bouts_length = {activity[0]: activity[1]['bout'] for activity in config.items()}
        df['activity'] = self._fix_activities_bouts(df, bouts_length)

        df['lie'] = self.get_lie(df, **config['lie'])
        df.loc[df['lie'], 'activity'] = 'lie'

        for activity in ['sit', 'lie']:
            df['activity'] = self.fix_bouts(df['activity'], activity, bouts_length[activity])

        df.loc[non_wear, 'activity'] = 'non-wear'
        del non_wear

        df['steps'] = self.get_steps(df)

        df.loc[(df['activity'] == 'walk') & (df['steps'] > config['run']['step_frequency']), 'activity'] = 'run'
        for activity in ['run', 'walk']:
            df['activity'] = self.fix_bouts(df['activity'], activity, bouts_length[activity])

        df.loc[df['activity'] == 'non-wear', 'direction'] = np.nan
        df.rename(
            columns={
                'inclination': 'thigh_inclination',
                'side_tilt': 'thigh_side_tilt',
                'direction': 'thigh_direction',
            },
            inplace=True,
        )

        references.update_angle(bouts, 'thigh')

        return df[['activity', 'steps', 'thigh_inclination', 'thigh_side_tilt', 'thigh_direction']]



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

import numpy as np
import pandas as pd

from .references import Angle, AngleStatus, Calibration

# NON-WEAR BOUTS SETTINGS
SHORT_OFF_BOUTS = 600
LONG_OFF_BOUTS = 5400
ON_BOUTS = 60
DEGREES_TOLERANCE = 5

logger = logging.getLogger(__name__)


class Calculation(Enum):
    AUTOMATIC = 'automatic'
    DEFAULT = 'default'


@dataclass
class Sensor(ABC):
    orientation: bool

    def get_angles(self, df: pd.DataFrame) -> pd.DataFrame:
        axes = df[['x', 'y', 'z']].to_numpy()

        euclidean_distance = np.linalg.norm(axes, axis=1)
        euclidean_distance = np.where(euclidean_distance == 0, np.nan, euclidean_distance)
        inclination = np.arccos(axes[:, 0] / euclidean_distance)
        side_tilt = -np.arcsin(axes[:, 1] / euclidean_distance)
        direction = -np.arcsin(axes[:, 2] / euclidean_distance)

        angles = np.column_stack((inclination, side_tilt, direction))
        angles = np.degrees(angles).astype(np.float32)

        return pd.DataFrame(
            angles,
            columns=['inclination', 'side_tilt', 'direction'],
            index=df.index,
        )

    def _get_small_bout_max(self, bout: pd.DataFrame, sd_sum: pd.Series) -> bool:
        start, end, _ = bout
        return sd_sum[start:end].max() > 0.5

    def _fix_off_bouts(self, df: pd.DataFrame) -> pd.Series:
        sd_mean = df[['sd_x', 'sd_y', 'sd_z']].mean(axis=1)
        sd_sum = df[['sd_x', 'sd_y', 'sd_z']].sum(axis=1)
        non_wear = sd_mean < 0.01
        non_wear.name = 'non-wear'

        bouts = (non_wear != non_wear.shift()).cumsum()
        bout_sizes = bouts[non_wear].value_counts()

        large_bouts = bout_sizes[bout_sizes > LONG_OFF_BOUTS].index.values
        small_bouts = bout_sizes[(bout_sizes <= LONG_OFF_BOUTS) & (bout_sizes > SHORT_OFF_BOUTS)].index.values

        if small_bouts.size > 0:
            small_bouts = bouts[bouts.isin(small_bouts)].drop_duplicates(keep='first').reset_index(drop=False)
            small_bouts.rename(columns={'datetime': 'end', 'non-wear': 'bout'}, inplace=True)
            small_bouts['end'] = small_bouts['end'] - pd.Timedelta(seconds=10)  # X seconds before start of short bout
            small_bouts.insert(
                0, 'start', small_bouts['end'] - pd.Timedelta(seconds=5)
            )  # Y seconds before end of short bout

            small_bouts['non-wear'] = small_bouts.apply(lambda x: self._get_small_bout_max(x, sd_sum), axis=1)
            small_bouts = small_bouts.loc[small_bouts['non-wear'], 'bout'].values

        non_wear = bouts.isin(large_bouts) | bouts.isin(small_bouts)

        return non_wear

    def _fix_on_bouts(
        self,
        non_wear: pd.Series,
    ) -> pd.Series:
        bouts = (non_wear != non_wear.shift()).cumsum()
        bout_sizes = bouts[~non_wear].value_counts()
        short_on_bouts = bout_sizes[bout_sizes < ON_BOUTS].index.values

        non_wear.loc[bouts.isin(short_on_bouts)] = True

        return non_wear

    def _get_angle_mean(self, df: pd.DataFrame) -> bool:
        mean = df.mean(axis=0)

        rule_1 = (abs(mean - [90, 0, 90]) < DEGREES_TOLERANCE).all().astype(bool)
        rule_2 = (abs(mean - [90, 0, -90]) < DEGREES_TOLERANCE).all().astype(bool)

        return rule_1 or rule_2

    def _fix_off_bouts_angles(
        self,
        non_wear: pd.Series,
        df: pd.DataFrame,
    ) -> pd.Series:
        bouts = (non_wear != non_wear.shift()).cumsum()
        bout_sizes = bouts[non_wear].value_counts()

        large_bouts = bout_sizes[bout_sizes > LONG_OFF_BOUTS].index.values
        other_bouts = bout_sizes[bout_sizes <= LONG_OFF_BOUTS].index.values

        other = bouts[bouts.isin(other_bouts)]

        other_bouts = df.groupby(other)[['inclination', 'side_tilt', 'direction']].apply(self._get_angle_mean)
        other_bouts = other_bouts[
            other_bouts
        ].index.values  # FIXME: Should this use FALSE or TRUE bouts from _get_angle_mean?

        non_wear = bouts.isin(large_bouts) | bouts.isin(other_bouts)

        return non_wear

    def get_non_wear(self, df: pd.DataFrame) -> pd.Series:
        non_wear = self._fix_off_bouts(df)
        non_wear = self._fix_on_bouts(non_wear)
        non_wear = self._fix_off_bouts_angles(non_wear, df)

        ratio = non_wear.value_counts(normalize=True).to_dict()
        total_time = (df.index[-1] - df.index[0]).floor('s')
        non_wear_ratio = ratio.get(True)

        if non_wear_ratio:
            non_wear_time = pd.Timedelta(seconds=non_wear_ratio * total_time.total_seconds()).floor('s')
            non_wear_percentage = non_wear_ratio * 100
        else:
            non_wear_time = pd.Timedelta(seconds=0)
            non_wear_percentage = 0

        logger.info(
            f'Non-wear detection: {non_wear_time} ({non_wear_percentage:.2f}%) out of {total_time} classified as non-wear time.'
        )

        return non_wear

    def fix_sensor_orientation(
        self,
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        df = df.copy()
        upside_down = self.check_upside_down_flip(df)
        inside_out = self.check_inside_out_flip(df)
        columns, text = None, None

        if upside_down and inside_out:
            columns = ['x', 'z', 'sum_x', 'sum_z', 'sum_dot_xz', 'direction']
            text = 'upside down and inside out'

        elif inside_out:
            columns = ['y', 'z', 'sum_y', 'sum_z', 'sum_dot_xz', 'side_tilt', 'direction']
            text = 'inside out'

        elif upside_down:
            columns = ['x', 'y', 'sum_x', 'sum_y', 'sum_dot_xz', 'side_tilt']
            text = 'upside down'

        if columns:
            df[columns] = -df[columns]
            df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)

            logger.warning(f'Sensor is {text}. Data flipped {columns} and angles recalculated.')
        else:
            logger.info('No sensor flip detected. No changes made to the data.')

        return df

    def check_upside_down_flip(self, df: pd.DataFrame) -> bool:
        valid_points = df[(df['inclination'] < 45) | (df['inclination'] > 135)]

        if valid_points.empty:
            logger.warning('Not enough data to check upside down flip. Skipping.')
            return False

        mdn = np.median(valid_points['x'])
        flip = True if (mdn < -0.1) else False

        if flip:
            logger.warning(f'Upside down flip detected (median x: {mdn:.2f}).')

        return flip

    @abstractmethod
    def check_inside_out_flip(self, df: pd.DataFrame) -> bool:
        pass

    @abstractmethod
    def rotate_by_reference_angle(self, df: pd.DataFrame, angle: float | np.ndarray) -> pd.DataFrame:
        pass

    @abstractmethod
    def calculate_reference_angle(self, df: pd.DataFrame) -> tuple[float | np.ndarray, Calculation]:
        pass

    def fix_bouts_orientation(self, df: pd.DataFrame, bouts: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()

        for start, end, non_wear, angle, calibration in bouts.to_numpy():
            if not non_wear:
                bout_df = df[start:end]
                df.loc[df.index.isin(bout_df.index)] = self.fix_sensor_orientation(bout_df)

        return df

    def rotate_bouts_by_reference_angles(
        self, df: pd.DataFrame, bouts: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        df = df.copy()
        bouts = bouts.copy()
        angles = []
        prev_angle_value = None

        for start, end, non_wear, calibration, angle in bouts.to_numpy():
            bout_df = df[start:end]

            if angle:
                angle.status = AngleStatus.PROPAGATED
                prev_angle_value = angle.value

            elif isinstance(calibration, Calibration):
                calibration_period = calibration.get_calibration_period(bout_df)
                value, status = self.calculate_reference_angle(calibration_period)
                status = AngleStatus.CALIBRATION if status == Calculation.AUTOMATIC else AngleStatus.DEFAULT
                expires = calibration.expires or None
                angle = Angle(value, expires, status)
                prev_angle_value = angle.value

            else:
                if non_wear and prev_angle_value:
                    value, expires, status = prev_angle_value, None, AngleStatus.DEFAULT
                else:
                    value, status = self.calculate_reference_angle(bout_df)
                    expires = None

                angle = Angle(value, expires, status)
                prev_angle_value = angle.value

            angles.append(angle)
            df.loc[bout_df.index] = self.rotate_by_reference_angle(bout_df, angle.value)

        bouts['angle'] = pd.Series(angles, dtype=object)

        return df, bouts

    def fix_bouts(
        self,
        activities: pd.Series,
        activity: str,
        bouts_length: int,
    ) -> pd.Series:
        df = activities.to_frame('activity')
        df['bout'] = (df['activity'] != df['activity'].shift()).cumsum()

        specific_activity = df.loc[df['activity'] == activity, 'bout']
        bout_sizes = specific_activity.value_counts()
        short_bouts = bout_sizes[bout_sizes <= bouts_length].index.values

        df['new_activity'] = df['activity']
        df.loc[df['bout'].isin(short_bouts), 'new_activity'] = np.nan

        half = len(short_bouts) // 2
        half = max(half, 1)

        df['activity'] = df['new_activity'].ffill(limit=half).bfill(limit=half)

        return df['activity']



from dataclasses import dataclass
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd


@dataclass
class FileParser:
    """Abstract base class for file parsers.

    Provides methods for checking file extensions and existence.
    """

    def check_extension(self, path: Path, extension: str) -> None:
        """Check if the file has the expected extension."""
        if isinstance(path, str):
            path = Path(path)

        if not path.suffix == extension:
            raise ValueError(f'Invalid file type: {path.suffix}. Expected {extension}.')

    def check_existence(self, path: Path) -> None:
        """Check if the file exists and is a file."""
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f'File not found: {path}.')

        if not path.is_file():
            raise ValueError(f'Path is not a file: {path}.')

    def check_file(self, path: Path, extension: str) -> None:
        """Check if the file exists and has the expected extension."""
        if isinstance(path, str):
            path = Path(path)

        self.check_existence(path)
        self.check_extension(path, extension)

    def check_empty(self, df: pd.DataFrame) -> None:
        """Check if the DataFrame is empty."""
        if not isinstance(df, pd.DataFrame):
            raise TypeError('Expected a pandas DataFrame.')

        if df.empty:
            raise ValueError('No data found in the file.')


SENS_NORMALIZATION_FACTOR = -4 / 512
DTYPE = np.dtype([('timestamp', '6uint8'), ('x', '>i2'), ('y', '>i2'), ('z', '>i2')])


@dataclass
class Sens(FileParser):
    normalize: bool = True

    def _read(
        self,
        obj: Path | bytes,
        func: Callable,
    ) -> pd.DataFrame:
        data = func(obj, dtype=DTYPE, count=-1, offset=0)
        timestamps = np.dot(data['timestamp'], [1 << 40, 1 << 32, 1 << 24, 1 << 16, 1 << 8, 1])

        df = pd.DataFrame(
            {
                'datetime': pd.to_datetime(timestamps, unit='ms', utc=True),
                'acc_x': data['x'].astype(np.int16),
                'acc_y': data['y'].astype(np.int16),
                'acc_z': data['z'].astype(np.int16),
            }
        )

        self.check_empty(df)

        df.set_index('datetime', inplace=True)

        if self.normalize:
            df = df * SENS_NORMALIZATION_FACTOR

        return df.astype(np.float32)

    def from_bin(self, path: str | Path) -> pd.DataFrame:
        if isinstance(path, str):
            path = Path(path)

        self.check_file(path, '.bin')

        return self._read(path, np.fromfile)

    def from_buffer(self, buffer: bytes) -> pd.DataFrame:
        if not isinstance(buffer, bytes):
            raise TypeError('Expected a bytes object.')

        return self._read(buffer, np.frombuffer)



import logging
import math
from dataclasses import dataclass
from datetime import timedelta
from typing import Any

import numpy as np
import pandas as pd
from numpy.fft import fft as np_fft
from numpy.lib.stride_tricks import sliding_window_view
from scipy import signal

from .calibration import AutoCalibrate
from .iterators import DataFrameIterator
from .settings import FEATURES, SENS__FLOAT_FACTOR, SENS__NORMALIZATION_FACTOR

logger = logging.getLogger(__name__)


@dataclass
class Features:
    """Processes raw accelerometer data to extract features.

    This class provides a pipeline for transforming raw accelerometer time-series
    data into a set of features. The process includes input validation,
    sampling frequency detection, resampling to a consistent frequency, optional
    auto-calibration, and the computation of various features like High-Low (HL)
    ratio, step-related metrics, and downsampled statistical summaries.

    The class can process data in a single batch or in overlapping chunks to
    mimic Sen's infrastructure processing.

    Attributes:
        system_frequency (int): The target frequency (in Hz) to which the data
            is resampled. Defaults to 30 Hz and should not be changed as all pipelines are
            designed to work with this frequency.
        validation (bool): Performs validation on the input DataFrame's
            format.
        calibrate (bool): Applies auto-calibration to the raw data.
        chunking (bool): Processes the data in chunks.
        size (timedelta): The size of each data chunk when chunking is enabled.
        overlap (timedelta): The overlap between consecutive chunks.
    """

    system_frequency: int = 30
    validation: bool = True
    calibrate: bool = True
    chunking: bool = False
    size: timedelta = timedelta(hours=24)
    overlap: timedelta = timedelta(minutes=15)

    def __post_init__(self):
        """Initializes the Features class after its construction.

        Converts size and overlap to timedelta objects if they are provided as strings.
        """
        if isinstance(self.size, str):
            self.size = pd.Timedelta(self.size).to_pytimedelta()

        if isinstance(self.overlap, str):
            self.overlap = pd.Timedelta(self.overlap).to_pytimedelta()

    def get_nyquist_freq(self, sampling_frequency: float) -> float:
        """Calculates the Nyquist frequency.

        Args:
            sampling_frequency: The sampling frequency.

        Returns:
            The Nyquist frequency.
        """
        return sampling_frequency / 2

    @staticmethod
    def get_sampling_frequency(
        df: pd.DataFrame, *, samples: int | None = 30_000, round_to_nearest: float | None = 0.5
    ) -> float:
        """Calculates the sampling frequency of a time-series DataFrame.
        This function determines the sampling frequency by finding the mode of the
        time differences between consecutive samples in the DataFrame's index.
        The calculation can be performed on a subset of the data for
        efficiency.

        Args:
            df (pd.DataFrame): The input DataFrame with a time-based index.
            samples (int | None, optional): The number of initial samples to use for
                the calculation. If None, the entire DataFrame is used.
            round_to_nearest (float | None, optional): The value to which the final
                frequency is rounded. For example, 0.5 will round to the nearest
                0.5 Hz. If None or non-positive, no rounding is performed.

        Returns:
            float: The estimated sampling frequency in Hertz (Hz).

        Raises:
            ValueError: If the DataFrame contains fewer than two samples, making
                frequency calculation impossible.
            ValueError: If the calculated time intervals are non-positive,
                indicating a non-monotonic or invalid time index.
        """
        time_subset = df.index[:samples] if samples else df.index

        if len(time_subset) < 2:
            raise ValueError('DataFrame must have at least 2 samples to calculate sampling frequency.')

        # Convert to nanoseconds then to seconds for time differences
        time_diffs_seconds = pd.Series(np.diff(time_subset.astype('int64')) / 1e9)

        sf = time_diffs_seconds.mode().values[0]

        if sf <= 0:
            raise ValueError('Invalid time intervals detected in data.')

        sf = 1.0 / sf
        if round_to_nearest and round_to_nearest > 0:
            sf = round(sf / round_to_nearest) * round_to_nearest

        logging.info(f'Detected sampling frequency: {sf:.2f} Hz.', extra={'sampling_frequency': sf})

        return sf

    def _resample_fft(self, df: pd.DataFrame) -> pd.DataFrame:
        """Resamples a DataFrame using the FFT method.

        Args:
            df: The input DataFrame.

        Returns:
            The resampled DataFrame.
        """
        start = df.index[0]
        end = df.index[-1]

        n_out = np.floor((end - start).total_seconds() * self.system_frequency).astype(int)
        resampled = signal.resample(df, n_out)

        df = pd.DataFrame(
            resampled,
            columns=df.columns,
            index=pd.date_range(start=start, end=end, periods=n_out),
            dtype=np.float32,
        )

        return df

    def resampling(self, df: pd.DataFrame, sampling_frequency: float, tolerance=1) -> pd.DataFrame:
        """Resamples a DataFrame to the system frequency if necessary.

        Args:
            df: The input DataFrame.
            sampling_frequency: The sampling frequency of the input DataFrame.
            tolerance: The tolerance for comparing sampling frequencies.

        Returns:
            The resampled DataFrame.
        """
        if math.isclose(sampling_frequency, self.system_frequency, abs_tol=tolerance):
            logger.info(
                f'Sampling frequency is {self.system_frequency} Hz, no resampling needed.',
                extra={'sampling_frequency': sampling_frequency},
            )
            return df

        df = self._resample_fft(df)

        return df

    def get_hl_ratio(self, df: pd.DataFrame) -> pd.Series:
        """Calculates the High-Low Ratio (HL-Ratio) from the Z-axis accelerometer data.
        This method processes the 'acc_z' signal by first separating it into
        high-frequency and low-frequency components using third-order low-pass and
        high-pass Butterworth filters with a cutoff frequency of 1 Hz.

        The absolute values of these filtered signals are then processed using a
        sliding window. The mean of the high-frequency components and the mean
        of the low-frequency components are calculated for each window. The
        HL-Ratio is the ratio of the mean of the high-frequency component to the
        mean of the low-frequency component. Division by zero is handled by
        returning zero for that window.

        Args:
            df (pd.DataFrame): Input DataFrame containing the time-series data.
                               Must include an 'acc_z' column for the Z-axis
                               accelerometer readings.

        Returns:
            pd.Series: A pandas Series named 'hl_ratio' containing the
                       calculated high-low ratio for each window.
        """

        order = 3
        cut_off = 1
        window = self.system_frequency * 4
        cut_off = cut_off / self.get_nyquist_freq(self.system_frequency)

        axis_z = df['acc_z'].values

        b, a = signal.butter(order, cut_off, 'low')
        low = signal.filtfilt(b, a, axis_z, axis=0)
        low = np.abs(low.astype(np.float32))

        b, a = signal.butter(order, cut_off, 'high')
        high = signal.filtfilt(b, a, axis_z, axis=0)
        high = np.abs(high.astype(np.float32))

        pad_width = window - 1
        high = np.pad(high, (0, pad_width), mode='edge')
        low = np.pad(low, (0, pad_width), mode='edge')

        high_windows = sliding_window_view(high, window)[:: self.system_frequency]
        mean_high = np.mean(high_windows, axis=1, dtype=np.float32)

        low_windows = sliding_window_view(low, window)[:: self.system_frequency]
        mean_low = np.mean(low_windows, axis=1, dtype=np.float32)

        hl_ratio = np.divide(
            mean_high, mean_low, out=np.zeros_like(mean_high), where=mean_low != 0
        )  # NOTE: Check what happens if mean_low is zero

        return pd.Series(hl_ratio, name='hl_ratio')

    def _get_steps_feature(self, arr: np.ndarray) -> np.ndarray:
        """Computes the steps feature from an array.

        Args:
            arr: The input array.

        Returns:
            An array containing the steps feature.
        """
        window = self.system_frequency * 4  # 120 (system frequency = 30) samples equal to 2 seconds
        steps_window = 4 * window  # 480 (system frequency = 30) samples equal to 8 seconds
        half_size = window * 2  # 240 (system frequency = 30) samples equal to 4 seconds
        arr = arr.astype(np.float32)

        pad_width = window - 1
        arr = np.pad(arr, (0, pad_width), mode='edge')

        windows = sliding_window_view(arr, window)[:: self.system_frequency]
        windows = windows - np.mean(windows, axis=1, keepdims=True, dtype=np.float32)

        fft_result = np_fft(windows, steps_window)[:, :half_size]
        magnitudes = 2 * np.abs(fft_result)

        return np.argmax(magnitudes, axis=1)

    def get_steps_features(self, df: pd.DataFrame) -> pd.DataFrame:
        """Calculates walking and running features from accelerometer data.
        This method processes the x-axis accelerometer signal ('acc_x') to
        extract features indicative of walking and running. It applies a
        series of Butterworth filters to isolate the frequency bands
        typically associated with these activities.

        The process involves:
        1. Applying a 6th-order low-pass filter with a 2.5 Hz cutoff.
        2. Applying a 6th-order high-pass filter with a 1.5 Hz cutoff to the
           result of step 1 to isolate the 'walk' signal.
        3. Applying a 6th-order high-pass filter with a 3 Hz cutoff to the
           'walk' signal to isolate the 'run' signal.
        4. Computing a feature value for both the 'walk' and 'run' signals
           using frequency component with the highest magnitude.

        Args:
            df (pd.DataFrame): The input DataFrame containing accelerometer data.
                               It must have an 'acc_x' column.

        Returns:
            pd.DataFrame: A new DataFrame with two columns: 'walk_feature' and
                          'run_feature'.
        """

        axis_x = df['acc_x'].values
        nyquist_frequency = self.get_nyquist_freq(self.system_frequency)

        b, a = signal.butter(6, 2.5 / nyquist_frequency, 'low')
        filtered = signal.lfilter(b, a, axis_x, axis=0)

        b, a = signal.butter(6, 1.5 / nyquist_frequency, 'high')
        walk = signal.lfilter(b, a, filtered, axis=0)

        b, a = signal.butter(6, 3 / nyquist_frequency, 'high')
        run = signal.lfilter(b, a, walk)

        df = pd.DataFrame(
            {
                'walk_feature': self._get_steps_feature(walk),
                'run_feature': self._get_steps_feature(run),
            },
        )

        return df

    def get_tensor(self, arr: np.ndarray) -> np.ndarray:
        """Creates a tensor from an array.

        Args:
            arr: The input array.

        Returns:
            A tensor.
        """
        pb = np.vstack((arr[: self.system_frequency], arr))
        pa = np.vstack((arr, arr[-self.system_frequency :]))
        n = pb.shape[0] // self.system_frequency
        tensor = np.concatenate(
            [
                pb[: n * self.system_frequency].reshape(self.system_frequency, n, 3, order='F'),
                pa[: n * self.system_frequency].reshape(self.system_frequency, n, 3, order='F'),
            ],
            axis=0,
        )
        return tensor[:, :-1, :]

    def downsampling(self, df: pd.DataFrame) -> pd.DataFrame:
        """Downsamples the input signal data and computes various statistical features.
        For each window, it calculates the mean, standard deviation, sum, and sum of squares for each axis
        (x, y, z), as well as the sum of the dot product between the x and z axes.

        Args:
            df (pd.DataFrame): The input DataFrame containing time-series signal
                data, expected to have columns representing different axes.

        Returns:
            pd.DataFrame: A new DataFrame where each row corresponds to a
                window of the original signal. The columns contain the
                computed features: 'x', 'y', 'z' (mean values), 'sd_x',
                'sd_y', 'sd_z', 'sum_x', 'sum_y', 'sum_z', 'sq_sum_x',
                'sq_sum_y', 'sq_sum_z', and 'sum_dot_xz'.
        """

        axes = df.values

        b, a = signal.butter(4, 5 / self.get_nyquist_freq(self.system_frequency), 'low')
        filtered = signal.lfilter(b, a, axes, axis=0).astype(np.float32)

        tensor = self.get_tensor(filtered)

        mean = np.mean(tensor, axis=0)
        sd = tensor.std(axis=0, ddof=1)
        sum = np.sum(tensor, axis=0)
        sq_sum = np.sum(np.square(tensor), axis=0)
        sum_dot_xz = np.sum((tensor[:, :, 0] * tensor[:, :, 2]), axis=0)

        df = np.concatenate([mean, sd, sum, sq_sum], axis=1)

        df = pd.DataFrame(
            df,
            columns=[
                'x',
                'y',
                'z',
                'sd_x',
                'sd_y',
                'sd_z',
                'sum_x',
                'sum_y',
                'sum_z',
                'sq_sum_x',
                'sq_sum_y',
                'sq_sum_z',
            ],
        )
        df['sum_dot_xz'] = sum_dot_xz

        return df

    def check_format(self, df: pd.DataFrame) -> pd.DataFrame:
        """Validates and standardizes the input accelerometer DataFrame.
        This method performs a series of checks to ensure the input DataFrame
        conforms to the expected format for accelerometer data.

        The validation ensures that the input is a non-empty pandas DataFrame with
        exactly three numeric columns and a datetime index. After validation, it
        renames the columns to ['acc_x', 'acc_y', 'acc_z'].

        Args:
            df (pd.DataFrame): The input DataFrame to validate. It is expected
                to have a datetime index and three numeric columns representing
                the x, y, and z axes of an accelerometer.

        Returns:
            pd.DataFrame: The validated and standardized DataFrame with columns
                renamed to ['acc_x', 'acc_y', 'acc_z'].

        Raises:
            TypeError: If the input is not a pandas DataFrame.
            ValueError: If the DataFrame is empty, does not have exactly 3
                columns, has a non-datetime index, or contains non-numeric
                column data.
        """

        if not self.validation:
            return df

        if not isinstance(df, pd.DataFrame):
            raise TypeError('Input must be a pandas DataFrame.')

        if df.empty:
            raise ValueError('DataFrame cannot be empty.')

        if df.shape[1] != 3:
            raise ValueError(
                f'DataFrame must have exactly 3 columns for accelerometer data, but has {df.shape[1]} columns.'
            )

        if not pd.api.types.is_datetime64_any_dtype(df.index):
            raise ValueError(f'DataFrame index must be of datetime type, but got {df.index.dtype}.')

        for col in df.columns:
            if not pd.api.types.is_numeric_dtype(df[col]):
                raise ValueError(f"Column '{col}' must contain numeric data, but got {df[col].dtype}.")

        df = df.iloc[:, :3]
        df.columns = ['acc_x', 'acc_y', 'acc_z']

        return df

    def _compute_chunk(
        self,
        df: pd.DataFrame,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Computes features for a single chunk of data.

        Args:
            df: The input DataFrame chunk.
            **kwargs: Additional keyword arguments.

        Returns:
            A DataFrame with computed features for the chunk.
        """
        not_overlaps = df[~df['overlap']]
        start, end = not_overlaps.index[0], not_overlaps.index[-1]

        df = self._compute(
            df.iloc[:, :3],
            **kwargs,
        )
        df = df.loc[(df.index >= start) & (df.index < end)]

        return df

    def _compute_chunks(self, df: pd.DataFrame, sampling_frequency: float) -> pd.DataFrame:
        """Computes features for a DataFrame in chunks.

        Args:
            df: The input DataFrame.
            sampling_frequency: The sampling frequency of the DataFrame.

        Returns:
            A DataFrame with computed features.
        """
        chunks = DataFrameIterator(df, size=self.size, overlap=self.overlap)
        computed = []

        for chunk in chunks:
            computed.append(self._compute_chunk(chunk, sampling_frequency=sampling_frequency))

        computed = pd.concat(computed)
        computed.sort_index(inplace=True)

        return computed

    def _compute(self, df: pd.DataFrame, sampling_frequency: float) -> pd.DataFrame:
        """Computes features for a DataFrame.

        Args:
            df: The input DataFrame.
            sampling_frequency: The sampling frequency of the DataFrame.

        Returns:
            A DataFrame with computed features.
        """
        df = self.resampling(df, sampling_frequency)
        hl_ratio = self.get_hl_ratio(df)
        steps_features = self.get_steps_features(df)
        downsampled = self.downsampling(df)

        n = min(len(hl_ratio), len(steps_features), len(downsampled))
        start = df.index[0].ceil('s')
        df = pd.concat([downsampled, hl_ratio, steps_features], axis=1)
        df = df.iloc[:n]
        df.index = pd.date_range(
            start=start,
            periods=n,
            freq=timedelta(seconds=1),
            name='datetime',
        )
        df['sf'] = sampling_frequency
        logger.info('Features computed.')

        return df

    def compute(self, df: pd.DataFrame, sampling_frequency: float | None = None) -> pd.DataFrame:
        """Computes features from the provided accelerometer data DataFrame.
        This method serves as the main entry point for feature computation. It handles
        pre-processing steps like data format validation, sampling frequency
        determination, and optional calibration before dispatching the computation
        to either a chunked or a non-chunked processing function based on the
        instance's configuration.

        Args:
            df (pd.DataFrame): The input DataFrame containing accelerometer data.
                It is expected to have a time-based index from which the sampling
                frequency can be inferred if not provided.

            sampling_frequency (float | None, optional): The sampling frequency of
                the data in Hertz. If None, it will be automatically calculated
                from the DataFrame's index.

        Returns:
            pd.DataFrame: A new DataFrame containing the computed features.
        """

        df = self.check_format(df)

        sampling_frequency = sampling_frequency or self.get_sampling_frequency(df)

        if self.calibrate:
            df = AutoCalibrate().compute(df, hertz=sampling_frequency)

        if self.chunking:
            return self._compute_chunks(df, sampling_frequency)
        else:
            return self._compute(df, sampling_frequency)

    def to_sens(
        self,
        df: pd.DataFrame,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Converts a DataFrame to the SENS format.

        Args:
            df: The input DataFrame.

        Returns:
            A tuple containing timestamps, data, features, and verbose arrays.
        """
        df = df.copy()
        df.index = df.index.astype(np.int64) // 10**6  # Time in milliseconds
        df.drop(columns=['sum_y', 'sq_sum_y'], inplace=True)

        df.fillna(0, inplace=True)
        df[FEATURES] = (df[FEATURES] * SENS__FLOAT_FACTOR).astype(np.int32)

        df['data'] = 1
        df['data'] = df['data'].astype(np.int16)

        df['verbose'] = 0
        df['verbose'] = df['verbose'].astype(np.int32)

        return (
            df.index.values,
            df['data'].values,
            df[FEATURES].values,
            df['verbose'].values,
        )

    def _raw_from_sens(
        self,
        timestamps: np.ndarray,
        data: np.ndarray,
    ) -> pd.DataFrame:
        """Converts raw SENS data to a DataFrame.

        Args:
            timestamps: An array of timestamps.
            data: An array of accelerometer data.

        Returns:
            A DataFrame created from the SENS data.
        """
        df = pd.DataFrame(
            data,
            index=timestamps,
            columns=['acc_x', 'acc_y', 'acc_z'],
        )

        df = df * SENS__NORMALIZATION_FACTOR
        df.index = pd.to_datetime(df.index, unit='ms')
        df.index.name = 'datetime'

        return df

    def _df_to_sens(
        self,
        df: pd.DataFrame,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
    ]:
        """Converts a DataFrame to the SENS format.

        Args:
            df: The input DataFrame.

        Returns:
            A tuple containing timestamps and data arrays.
        """
        df = df / SENS__NORMALIZATION_FACTOR
        timestamps = (df.index.astype(np.int64) // 10**6).values
        timestamps = np.array([timestamps])

        data = df.values
        data = np.array([data])

        return timestamps, data

    def compute_sens(
        self,
        timestamps: np.ndarray,
        data: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Computes features from SENS data.

        Args:
            timestamps: An array of timestamps.
            data: An array of accelerometer data.

        Returns:
            A tuple containing timestamps, data, features, and verbose arrays in the SENS format.
        """
        df = self._raw_from_sens(timestamps[0], data[0])
        features = self.compute(df)

        return self.to_sens(features)



import logging
import warnings
from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
from skdh.preprocessing import CalibrateAccelerometer

logger = logging.getLogger(__name__)


@dataclass
class AutoCalibrate:
    min_hours: int = 24

    def compute(
        self,
        df: pd.DataFrame,
        hertz: float,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Performs accelerometer calibration on the input data.
        This method uses the `CalibrateAccelerometer` class from `skdh.preprocessing` library to correct for biases
        and scaling errors in raw accelerometer signals. It prepares the data,
        runs the calibration, and returns a new DataFrame with the calibrated values.
        If the calibration process fails, it logs a warning and returns the
        original, uncalibrated data.

        Args:
            df (pd.DataFrame): The input DataFrame containing raw accelerometer data.
                The index should be a datetime-like object representing time, and
                the columns should represent the accelerometer axes (e.g., 'x', 'y', 'z').
            hertz (float): The sampling frequency of the accelerometer data in Hertz.
            **kwargs (Any): Additional keyword arguments to be passed to the
                `CalibrateAccelerometer` initializer.

        Returns:
            pd.DataFrame: A DataFrame with the same shape and index as the input,
                containing the calibrated accelerometer data as float32 type. If
                calibration fails, the original data is returned.
        """

        columns = df.columns

        # Prepare data for calibration
        time = df.index
        accel = df.values
        del df

        # Initialize calibrator
        calibrator = CalibrateAccelerometer(min_hours=self.min_hours, **kwargs)

        # Perform calibration
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            calibrated = calibrator.predict(
                time=(time.astype(np.int64) // 10**9).values,
                accel=accel,
                fs=hertz,
            ).get('accel')

        if calibrated is None:
            logger.warning('Calibration did not produce valid results. Returning original accelerometer data.')
        else:
            logger.info('Calibration completed successfully.')
            accel = calibrated
            del calibrated

        # Create result DataFrame with calibrated accelerometer data
        return pd.DataFrame(accel, columns=columns, index=time, dtype=np.float32)



from dataclasses import dataclass
from datetime import timedelta

import pandas as pd

from .settings import ACTIVITIES


@dataclass
class Exposures:
    window: str | timedelta = '1d'

    def __post_init__(self):
        if isinstance(self.window, str):
            self.window = pd.Timedelta(self.window).to_pytimedelta()

    def _get_exposure(self, df: pd.DataFrame, valid: pd.Series, function: str) -> pd.Timedelta | int:
        if function == 'time':
            result = df.loc[valid, 'activity'].count()
            result = pd.Timedelta(result, unit='s')
        elif function == 'count':
            transitions = valid & ~(valid.shift(-1, fill_value=False))
            result = transitions.sum()
        else:
            raise ValueError(f'Unknown function: {function}')

        return result

    def get_bending(
        self,
        df: pd.DataFrame,
        lower: int,
        upper: int,
    ) -> pd.Series:
        valid = (
            df['activity'].isin(['stand', 'shuffle', 'walk', 'run', 'stairs'])
            & (df['trunk_direction'] > 0)
            & (df['trunk_inclination'].between(lower, upper, inclusive='both'))
        )

        return valid

    def get_arm_lifting(
        self,
        df: pd.DataFrame,
        lower: int,
        upper: int,
    ) -> pd.Series:
        valid = df['activity'].isin(['stand', 'shuffle', 'walk']) & (
            df['arm_inclination'].between(lower, upper, inclusive='both')
        )

        return valid

    def get_fast_walking(self, df: pd.DataFrame) -> pd.Series:
        return (df['activity'] == 'walk') & (df['steps'] >= 120 / 60)

    def get_slow_walking(self, df: pd.DataFrame) -> pd.Series:
        return (df['activity'] == 'walk') & (df['steps'] < 120 / 60)

    def _get_exposures(self, df: pd.DataFrame) -> pd.Series:
        exposure = {
            'wear': self._get_exposure(df, df['activity'] != 'non-wear', 'time'),
            'sedentary': self._get_exposure(df, df['activity'].isin(['sit', 'lie']), 'time'),
            'standing': self._get_exposure(df, df['activity'].isin(['stand', 'shuffle']), 'time'),
            'on_feet': self._get_exposure(
                df, df['activity'].isin(['stand', 'shuffle', 'walk', 'runk', 'stairs']), 'time'
            ),
            'sedentary_to_other': self._get_exposure(df, df['activity'].isin(['sit', 'lie']), 'count'),
            'lpa': self._get_exposure(
                df,
                df['activity'].isin(['stand', 'shuffle']) | self.get_slow_walking(df),
                'time',
            ),
            'mvpa': self._get_exposure(
                df,
                df['activity'].isin(['run', 'stairs', 'bicycle', 'row']) | self.get_fast_walking(df),
                'time',
            ),
        }

        if ('trunk_direction' in df.columns) and ('trunk_inclination' in df.columns):
            exposure['bending_30_60'] = self._get_exposure(df, self.get_bending(df, 30, 60), 'time')
            exposure['bending_60_90'] = self._get_exposure(df, self.get_bending(df, 60, 180), 'time')
            exposure['bending_45_180'] = self._get_exposure(df, self.get_bending(df, 45, 180), 'count')

        if 'arm_inclination' in df.columns:
            exposure['arm_lifting_30_60'] = self._get_exposure(df, self.get_arm_lifting(df, 30, 60), 'time')
            exposure['arm_lifting_60_90'] = self._get_exposure(df, self.get_arm_lifting(df, 60, 90), 'time')
            exposure['arm_lifting_90_180'] = self._get_exposure(df, self.get_arm_lifting(df, 90, 180), 'time')

        exposure = pd.Series(exposure)

        return exposure

    def _get_activities(self, df: pd.DataFrame) -> pd.DataFrame:
        columns = ACTIVITIES.values()
        columns = [col for col in columns if col in df['activity'].cat.categories]
        columns.remove('non-wear')

        activities = (
            df['activity'].groupby([pd.Grouper(freq=self.window, sort=True), df['activity']], observed=False).count()  # type: ignore
        )
        activities = activities.apply(pd.Timedelta, unit='s').unstack()
        return activities[columns]

    def compute(self, df: pd.DataFrame, activities: bool = False) -> pd.DataFrame:
        exposure = df.groupby(pd.Grouper(freq=self.window, sort=True)).apply(self._get_exposures)  # type: ignore

        if activities:
            df = self._get_activities(df)
            exposure = pd.concat([exposure, df], axis=1)

        return exposure



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Literal, Self

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

SECOND = timedelta(seconds=1)
MIN_CALIBRATION_SECONDS = 10


class AngleStatus(Enum):
    PROPAGATED = 'propagated'
    CALIBRATION = 'calibration'
    DEFAULT = 'default'


@dataclass
class Angle:
    value: float | list[float] | np.ndarray
    expires: datetime | None = None
    status: AngleStatus | None = None

    def __post_init__(self) -> Self:
        """Convert expires to appropriate types."""
        self.expires = pd.to_datetime(self.expires) if self.expires else None

        return self

    def _updated_bouts_by_angle(self, bouts: pd.DataFrame) -> pd.DataFrame:
        if not self.expires:
            return bouts

        updated_bouts = []
        angle = self

        for start, end, non_wear in bouts.to_numpy():
            if non_wear:
                angle = None
                updated_bouts.append({'start': start, 'end': end, 'non_wear': True, 'angle': angle})

            else:
                # If the angle expires in the bout, split the bout, else keep it as is.
                if self.expires > start and self.expires < end:
                    updated_bouts.append({'start': start, 'end': self.expires, 'non_wear': False, 'angle': angle})
                    angle = None
                    updated_bouts.append(
                        {'start': self.expires + SECOND, 'end': end, 'non_wear': False, 'angle': angle}
                    )

                else:
                    updated_bouts.append({'start': start, 'end': end, 'non_wear': False, 'angle': angle})

        return pd.DataFrame(updated_bouts)

    def is_expired(self, date: datetime) -> bool:
        """Check if the angle is expired based on the given date."""
        if self.expires is None:
            return False
        return self.expires < date

    def fix_expiration(self, calibrations: list['Calibration']) -> None:
        for calibration in calibrations:
            if self.expires and calibration.start < self.expires:
                self.expires = calibration.start - SECOND
                logger.info(f'Angle expiration adjusted to {self.expires} due to calibration overlap.')
                break


@dataclass
class Calibration:
    start: datetime
    end: datetime
    ttl: timedelta | None = None
    expires: datetime | None = None

    def __post_init__(self):
        """Convert start, end, ttl, and expires to appropriate types."""
        self.start = pd.to_datetime(self.start)
        self.end = pd.to_datetime(self.end)
        self.ttl = pd.Timedelta(self.ttl) if self.ttl else None

        self.expires = pd.to_datetime(self.expires) if self.expires else None
        self.expires = self.end + self.ttl if self.ttl and not self.expires else self.expires

    def is_outdated(self, date: datetime) -> bool:
        """Check if the calibration is outdated based on the given date."""
        return self.end < date

    def get_overlap_in_seconds(self, start: datetime, end: datetime):
        """Calculate the overlap in seconds between the calibration and a given interval."""
        overlap_start = max(self.start, start)
        overlap_end = min(self.end, end)

        if overlap_start < overlap_end:
            return (overlap_end - overlap_start).total_seconds()
        return 0

    def get_calibration_period(self, df: pd.DataFrame) -> pd.DataFrame:
        return df[self.start : self.end]


@dataclass
class References:
    calibrations: list[Calibration] | None = None
    thigh: Angle | None = None
    trunk: Angle | None = None

    def __post_init__(self) -> Self:
        """Initialize the References object."""
        if self.calibrations:
            self.calibrations = [Calibration(**cal) if isinstance(cal, dict) else cal for cal in self.calibrations]
            self.sort_calibrations()
            self.fix_calibration_overlaps()

        """Initialize the Angles object."""
        if isinstance(self.thigh, dict):
            self.thigh = Angle(**self.thigh)

            if self.calibrations:
                self.thigh.fix_expiration(self.calibrations)

        if isinstance(self.trunk, dict):
            self.trunk = Angle(**self.trunk)

            if self.calibrations:
                self.trunk.fix_expiration(self.calibrations)

        return self

    @classmethod
    def from_dict(cls, references: dict[str, Any] | None) -> Self:
        """Create a References object from a dictionary."""
        if references is None:
            logger.info('No references provided, returning empty References object.')
            return cls()

        return cls(**references)

    def to_dict(self) -> dict[str, Any]:
        """Convert the References object to a dictionary."""
        references = {}

        if self.calibrations:
            references['calibrations'] = [cal.__dict__ for cal in self.calibrations]
        if self.thigh:
            references['thigh'] = self.thigh.__dict__
        if self.trunk:
            references['trunk'] = self.trunk.__dict__

        return references

    def sort_calibrations(self):
        """Sort calibrations by start time."""
        if self.calibrations:
            self.calibrations.sort(key=lambda x: x.start)

    def fix_calibration_overlaps(self):
        """Fix overlapping calibration intervals."""
        second = timedelta(seconds=1)
        for index, interval in enumerate(self.calibrations):
            next_interval = self.calibrations[index + 1] if index + 1 < len(self.calibrations) else None

            if next_interval:
                if interval.expires > next_interval.start:
                    interval.expires = next_interval.start - second

    def remove_outdated_calibrations(self, date: datetime) -> None:
        """Remove calibrations that are outdated based on the dataframe start time."""
        if self.calibrations:
            for calibration in self.calibrations:
                if calibration.is_outdated(date):
                    logger.info(
                        f'Calibration {calibration.start} - {calibration.end} is outdated and has been removed.'
                    )
                    self.calibrations.remove(calibration)

    def remove_expired_angles(self, date: datetime) -> None:
        """Remove angles that have expired based on the dataframe start time."""
        if self.thigh and self.thigh.is_expired(date):
            self.thigh = None
            logger.info('Thigh angle reference has expired and has been removed.')

        if self.trunk and self.trunk.is_expired(date):
            self.trunk = None
            logger.info('Trunk angle reference has expired and has been removed.')

    def remove_outdated(self, date: datetime) -> None:
        """Remove outdated references based on the dataframe start time."""
        self.remove_outdated_calibrations(date)
        self.remove_expired_angles(date)

    def _update_bout_by_calibrations(self, start: pd.Timestamp, end: pd.Timestamp) -> list[dict[str, Any]]:
        calibration_bouts = []
        for calibration in self.calibrations:
            # Check if the calibration period is within the wear bout.
            if calibration.get_overlap_in_seconds(start, end) > MIN_CALIBRATION_SECONDS:
                bout_start, bout_end = start, end

                # If calibration starts after the wear bout starts, adjust the start time.
                if calibration.start > start:
                    bout_start = calibration.start

                # If the calibration angle expires before the wear bout ends, adjust the end time.
                if calibration.expires and calibration.expires < end:
                    bout_end = calibration.expires

                calibration_bouts.append(
                    {
                        'start': bout_start,
                        'end': bout_end,
                        'non_wear': False,
                        'calibration': calibration,
                    }
                )

        return calibration_bouts

    def _fill_bout_gaps(
        self,
        calibration_bouts: list[dict[str, Any]],
        start: pd.Timestamp,
        end: pd.Timestamp,
        angle: Angle | None,
    ) -> list[dict[str, Any]]:
        new_bouts = []
        new_start = start
        for calibration_bout in calibration_bouts:
            if new_start < calibration_bout['start']:
                new_bouts.append(
                    {
                        'start': new_start,
                        'end': calibration_bout['start'] - SECOND,
                        'non_wear': False,
                        'angle': angle,
                    }
                )

            new_start = calibration_bout['end'] + SECOND

        if new_start < end:
            new_bouts.append(
                {
                    'start': new_start,
                    'end': end,
                    'non_wear': False,
                }
            )

        new_bouts.extend(calibration_bouts)
        new_bouts.sort(key=lambda x: x['start'])

        return new_bouts

    def _updated_bouts_with_calibrations(self, bouts: pd.DataFrame) -> pd.DataFrame:
        if 'angle' not in bouts.columns:
            bouts['angle'] = None

        updated_bouts = []

        for start, end, non_wear, angle in bouts.to_numpy():
            if non_wear:
                updated_bouts.append({'start': start, 'end': end, 'non_wear': True, 'angle': angle})

            else:
                if self.calibrations:
                    calibration_bouts = self._update_bout_by_calibrations(start, end)
                    if calibration_bouts:
                        new_bouts = self._fill_bout_gaps(calibration_bouts, start, end, angle)
                        updated_bouts.extend(new_bouts)
                    else:
                        updated_bouts.append({'start': start, 'end': end, 'non_wear': False, 'angle': angle})
                else:
                    updated_bouts.append({'start': start, 'end': end, 'non_wear': False, 'angle': angle})

        updated_bouts = pd.DataFrame(updated_bouts)

        return updated_bouts

    def _get_non_wear_bouts(self, non_wear: pd.Series) -> pd.DataFrame:
        bouts = (non_wear != non_wear.shift()).cumsum().to_frame(name='bouts')
        bouts['non_wear'] = non_wear
        bouts = (
            bouts.reset_index()
            .groupby('bouts')
            .agg(
                start=('datetime', 'first'),
                end=('datetime', 'last'),
                non_wear=('non_wear', 'first'),
            )
        )

        return bouts

    def get_bouts(self, non_wear: pd.Series, angle: Literal['thigh', 'trunk'] | None = None) -> pd.DataFrame:
        bouts = self._get_non_wear_bouts(non_wear)

        if angle:
            angle = getattr(self, angle, None)
            bouts = angle._updated_bouts_by_angle(bouts) if isinstance(angle, Angle) else bouts

        bouts = self._updated_bouts_with_calibrations(bouts) if self.calibrations else bouts

        for col in ['calibration', 'angle']:
            if col not in bouts.columns:
                bouts[col] = None

            bouts[col] = bouts[col].replace(np.nan, None)

        return bouts[['start', 'end', 'non_wear', 'calibration', 'angle']]

    def update_angle(self, df: pd.DataFrame, angle: Literal['thigh', 'trunk']) -> None:
        last_angle = df['angle'].values[-1]

        if not isinstance(last_angle, Angle) or (
            last_angle.status not in [AngleStatus.PROPAGATED, AngleStatus.CALIBRATION]
        ):
            last_angle = None

        setattr(self, angle, last_angle)



import logging
from dataclasses import dataclass

import numpy as np
import pandas as pd

from .references import References
from .sensor import Sensor

logger = logging.getLogger(__name__)


@dataclass
class Arm(Sensor):
    def compute_activities(
        self,
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        df = df.copy()

        references = References()

        df[['inclination', 'side_tilt', 'direction']] = self.get_angles(df)

        # TODO: Is non-wear detection correct for arm? Produces non-wear times even if the arm is worn most probably.
        df['non-wear'] = self.get_non_wear(df)
        bouts = references.get_bouts(df['non-wear'])

        if self.orientation:
            df = self.fix_bouts_orientation(df, bouts)

        df.loc[df['non-wear'], 'inclination'] = np.nan
        df.rename(columns={'inclination': 'arm_inclination'}, inplace=True)

        return pd.DataFrame(df['arm_inclination'])

    def check_inside_out_flip(self, df: pd.DataFrame) -> bool:
        return False

    def rotate_by_reference_angle(self):
        raise NotImplementedError('Calf does not support rotating by reference angle.')

    def calculate_reference_angle(self):
        raise NotImplementedError('Calf does not support calculating reference angles.')



from datetime import timedelta

import pandas as pd


class DataFrameIterator:
    def __init__(
        self,
        df: pd.DataFrame,
        size: str | timedelta = '1d',
        overlap: str | timedelta = '15min',
    ) -> None:
        size = pd.Timedelta(size).to_pytimedelta()
        overlap = pd.Timedelta(overlap).to_pytimedelta()

        self._index = 0
        self.df = df
        self.size = size
        self.overlap = overlap
        self.chunks = self._get_chunks(df, size, overlap)

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self.chunks):
            chunk = self._load_chunk(self._index)

            self._index += 1
            return chunk

        else:
            raise StopIteration

    def _get_chunks(
        self,
        df: pd.DataFrame,
        size: timedelta,
        overlap: timedelta,
    ) -> pd.DataFrame:
        start = df.index.min()
        end = df.index.max()

        chunks = pd.date_range(start=start, end=end, freq=size, normalize=False).to_frame(index=False, name='start')
        chunks.index.name = 'chunk'
        chunks['end'] = chunks['start'] + size
        chunks.iat[0, 0] = start
        chunks.iat[-1, -1] = end

        chunks['start_overlap'] = chunks['start'] - overlap  # type: ignore
        chunks['end_overlap'] = chunks['end'] + overlap

        return chunks

    def _load_chunk(self, chunk: int) -> pd.DataFrame:
        if len(self.chunks) == 1:
            df = self.df.copy()
            df['overlap'] = False

            return df

        chunk_record = self.chunks.iloc[chunk]
        start, end = chunk_record['start'], chunk_record['end']
        start_overlap, end_overlap = chunk_record['start_overlap'], chunk_record['end_overlap']

        df = self.df  # type: pd.DataFrame
        df = df.loc[(df.index >= start_overlap) & (df.index < end_overlap)].copy()
        df['overlap'] = True
        df.loc[(df.index >= start) & (df.index < end), 'overlap'] = False

        return df



LEGACY_CONFIG = {
    'thigh': {
        'sit': {
            'bout': 5,
            'inclination_angle': 45,
        },
        'lie': {
            'bout': 1,
            'orientation_angle': 65,
        },
        'stand': {
            'bout': 2,
            'inclination_angle': 45,
            'movement_threshold': 0.1,
        },
        'walk': {
            'bout': 2,
            'inclination_angle': 45,
            'movement_threshold': 0.1,
            'run_threshold': 0.72,
        },
        'stairs': {
            'bout': 5,
            'inclination_angle': 45,
            'movement_threshold': 0.1,
            'run_threshold': 0.72,
            'direction_threshold': 40,
            'stairs_threshold': 4,
            'anterior_posterior_angle': 25,
        },
        'run': {
            'bout': 2,
            'inclination_angle': 45,
            'run_threshold': 0.72,
            'step_frequency': 2.5,
        },
        'bicycle': {
            'bout': 15,
            'movement_threshold': 0.1,
            'anterior_posterior_angle': 25,
            'direction_threshold': 40,
            'inclination_angle': 90,
        },
        'row': {
            'bout': 15,
            'movement_threshold': 0.1,
            'inclination_angle': 90,
        },
        'shuffle': {
            'bout': 2,
        },
    },
    'trunk': {
        'lie': {
            'inclination_angle': 45,
            'orientation_angle': 65,
        }
    },
}

CONFIG = {
    'thigh': {
        'sit': {
            'bout': 5,
            'inclination_angle': 47.5,
        },
        'lie': {
            'bout': 1,
            'orientation_angle': 65,
        },
        'stand': {
            'bout': 2,
            'inclination_angle': 47.5,
            'movement_threshold': 0.075,
        },
        'walk': {
            'bout': 2,
            'inclination_angle': 47.5,
            'movement_threshold': 0.075,
            'run_threshold': 0.7,
        },
        'stairs': {
            'bout': 5,
            'inclination_angle': 47.5,
            'movement_threshold': 0.075,
            'run_threshold': 0.7,
            'direction_threshold': 35.0,
            'stairs_threshold': 5,
            'anterior_posterior_angle': 20,
        },
        'run': {
            'bout': 2,
            'inclination_angle': 47.5,
            'run_threshold': 0.7,
            'step_frequency': 2.5,
        },
        'bicycle': {
            'bout': 15,
            'movement_threshold': 0.075,
            'anterior_posterior_angle': 20,
            'direction_threshold': 35.0,
            'inclination_angle': 87.5,
        },
        'row': {
            'bout': 15,
            'movement_threshold': 0.075,
            'inclination_angle': 87.5,
        },
        'shuffle': {
            'bout': 2,
        },
    },
    'trunk': {
        'lie': {
            'inclination_angle': 47.5,
            'orientation_angle': 65,
        }
    },
}

ACTIVITIES = {
    0: 'non-wear',
    1: 'lie',
    2: 'sit',
    3: 'stand',
    4: 'shuffle',
    5: 'walk',
    6: 'run',
    7: 'stairs',
    8: 'bicycle',
    9: 'row',
    10: 'kneel',
    11: 'squat',
}

FEATURES = [
    'x',
    'y',
    'z',
    'sd_x',
    'sd_y',
    'sd_z',
    'sum_x',
    'sum_z',
    'sq_sum_x',
    'sq_sum_z',
    'sum_dot_xz',
    'hl_ratio',
    'walk_feature',
    'run_feature',
    'sf',
]

# Sens backend specific settings
SENS__FLOAT_FACTOR = 1_000_000
SENS__NORMALIZATION_FACTOR = -4 / 512

SENS__ACTIVITY_VALUES = [
    'steps',
    'trunk_inclination',
    'trunk_side_tilt',
    'trunk_direction',
    'arm_inclination',
]  # "activity" is always present in the dataframe
